package ru.dbotthepony.mc.otm.worldgen.placement

import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.FloatProvider
import net.minecraft.util.valueproviders.IntProvider
import net.minecraft.world.level.levelgen.placement.PlacementModifierType
import ru.dbotthepony.mc.otm.registry.data.MPlacementModifiers
import java.util.stream.Stream

/**
 * Enormous ellipsoid ("cloud") placement, suitable to be used as ONLY primary placement modifier
 *
 * This placement modifier is designed to be terminal; other placement modifiers after this MUST NOT be placed
 *
 * @see AbstractEnormousPlacement
 */
class EnormousEllipsoidPlacement(
	parameters: Parameters,
	override val x: FloatProvider,
	override val z: FloatProvider,
	override val y: FloatProvider,
	override val count: IntProvider,
	override val xLength: FloatProvider,
	override val zLength: FloatProvider,
	override val yLength: FloatProvider,
) : AbstractEnormousPlacement(parameters), IEllipsoidPlacement {
	init {
		require(xLength.minValue >= 1f) { "Bad ellipsoid x minimal size: $xLength" }
		require(zLength.minValue >= 1f) { "Bad ellipsoid z minimal size: $zLength" }
		require(yLength.minValue >= 1f) { "Bad ellipsoid y minimal size: $yLength" }
	}

	override fun type(): PlacementModifierType<*> {
		return MPlacementModifiers.ENORMOUS_ELLIPSOID_PLACEMENT
	}

	override fun getPositions(center: BlockPos, random: RandomSource): Stream<BlockPos> {
		return getEllipsoidPositions(random, center)
	}

	companion object {
		val CODEC: MapCodec<EnormousEllipsoidPlacement> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					PARAMETERS_CODEC.forGetter(EnormousEllipsoidPlacement::parameters),
					FloatProvider.CODEC.fieldOf("x").forGetter(EnormousEllipsoidPlacement::x),
					FloatProvider.CODEC.fieldOf("y").forGetter(EnormousEllipsoidPlacement::y),
					FloatProvider.CODEC.fieldOf("z").forGetter(EnormousEllipsoidPlacement::z),
					IntProvider.CODEC.fieldOf("count").forGetter(EnormousEllipsoidPlacement::count),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("x_size").forGetter(EnormousEllipsoidPlacement::xLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("z_size").forGetter(EnormousEllipsoidPlacement::zLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("y_size").forGetter(EnormousEllipsoidPlacement::yLength),
				).apply(it, ::EnormousEllipsoidPlacement)
			}
		}
	}
}
